from pyticketswitch.event import Event
from pyticketswitch.performance import Performance
from pyticketswitch.seat import Seat
from pyticketswitch.send_method import SendMethod
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.commission import Commission


class TicketOrder(JSONMixin, object):
    """Describes a set of identical priced tickets.

    Attributes:
        code (str): the discount code.
        seats (list): list of seat IDs.
        number_of_seats (int): number of seats of this type the ticket
            order represents.
        description (str): the discount description.
        seatprice (float): price of individual seat in this ticket order.
        surcharge (float): additional surcharge added per seat in this
            ticket order.
        total_seatprice (float): seatprice for all tickets in the ticket order.
        total_surcharge (float): surcharge for all tickets in the ticket order.

    """

    def __init__(self, code, seats=None, number_of_seats=None, description=None,
                 seatprice=None, surcharge=None, total_seatprice=None,
                 total_surcharge=None):
        self.code = code
        self.seats = seats
        self.description = description
        self.number_of_seats = number_of_seats
        self.seatprice = seatprice
        self.surcharge = surcharge
        self.total_seatprice = total_seatprice
        self.total_surcharge = total_surcharge

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **TicketOrder** object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a ticket order.

        Returns:
            :class:`TicketOrder <pyticketswitch.order.TicketOrder>`: a new
            :class:`TicketOrder <pyticketswitch.order.TicketOrder>` object
            populated with the data from the api.

        """

        kwargs = {
            'code': data.get('discount_code'),
            'number_of_seats': data.get('no_of_seats'),
            'description': data.get('discount_desc'),
        }

        # Below we are explicital checking for not None because we want to
        # differentiate between situtations where a value is 0 and a value is
        # missing from the response.
        raw_seatprice = data.get('sale_seatprice')
        if raw_seatprice is not None:
            kwargs.update(seatprice=raw_seatprice)

        raw_surcharge = data.get('sale_surcharge')
        if raw_surcharge is not None:
            kwargs.update(surcharge=raw_surcharge)

        raw_total_seatprice = data.get('total_sale_seatprice')
        if raw_total_seatprice is not None:
            kwargs.update(total_seatprice=raw_total_seatprice)

        raw_total_surcharge = data.get('total_sale_surcharge')
        if raw_total_surcharge is not None:
            kwargs.update(total_surcharge=raw_total_surcharge)

        seats_data = data.get('seats')
        if seats_data:
            seats = [
                Seat.from_api_data(seat)
                for seat in seats_data
            ]
            kwargs.update(seats=seats)

        return cls(**kwargs)

    def combined_price(self):
        """Returns the combined seatprice and surcharge.

        This method assumes that we have both a seatprice and surcharge.
        In the situation where are missing either a seatprice or a surcharge
        then we don't have all the information to be able provide this
        information.

        Returns:
            float: the ticket order seatprice and surcharge

        Raises:
            AssertionError: It might seem like the obvious thing to do would be
                to assume the missing data was in fact zero and simply allow the
                addition to continue. However that would be somewhat dangerous when
                we are talking about prices, and it's better to actually raise an
                exception to indicate that there was a problem with the objects
                data, than to inform a customer that the tickets are free or have
                no booking fees

        """
        assert self.seatprice is not None, 'seatprice data missing'
        assert self.surcharge is not None, 'surcharge data missing'
        return self.seatprice + self.surcharge

    def total_combined_price(self):
        """Returns the combined total seatprice and surcharge.

        This method assumes that we have both a seatprice and surcharge.
        In the situation where are missing either a seatprice or a surcharge
        then we don't have all the information to be able provide this
        information.

        Returns:
            float: the ticket order total seatprice and total surcharge

        Raises:
            AssertionError: It might seem like the obvious thing to do would be
                to assume the missing data was in fact zero and simply allow the
                addition to continue. However that would be somewhat dangerous when
                we are talking about prices, and it's better to actually raise an
                exception to indicate that there was a problem with the objects
                data, than to inform a customer that the tickets are free or have
                no booking fees

        """
        assert self.total_seatprice is not None, 'seatprice data missing'
        assert self.total_surcharge is not None, 'surcharge data missing'
        return self.total_seatprice + self.total_surcharge

    def __repr__(self):
        return u'<TicketOrder {}>'.format(self.code)


class Order(JSONMixin, object):
    """Describes tickets for a specific event/performance.

    Attributes:
        item (int): the item number within the trolley. If an order is removed
            from the trolley it will retain it's item number.
        event (:class:`Event <pyticketswitch.event.Event>`): the event
            the order is for.
        performance (:class:`Performance <pyticketswitch.performance.Performance>`):
            the performance the order is for.
        price_band_code (str): the price band identifier.
        price_band_description (str): the optional description for the price band
        ticket_type_code (str): the ticket type identifier.
        ticket_type_description (str): description of the ticket type.
        ticket_orders (list): list of
            :class:`TicketOrders <pyticketswitch.order.TicketOrder>` that
            provide information on the specific tickets in the order.
        number_of_seats (int): total number of seats/tickets in the
            order.
        total_seatprice (float): total seat price of the order.
        total_surcharge (float): total additional surcharges for the order.
        seat_request_status (str): the status of requested seats. Will be filled
            after the order has been reserved, and :obj:`None` before.
        requested_seats (list): list of
            :class:`Seats <pyticketswitch.seat.Seat>` requested at reservation
            time.
        backend_purchase_reference (str): a reference from the source
            supplier for this order. This is generally empty until the order
            has been successfully purchased.
        send_method (:class:`SendMethod <pyticketswitch.send_method.SendMethod>`):
            method of ticket delivery. Only present when requested.
        gross_commission (:class: `Commission <pyticketswitch.commission.Commission>`):
            predicted commission to be shared between ingresso and the partner.
        user_commission (:class: `Commission <pyticketswitch.commission.Commission>`):
            predicted commission for the partner.

    """

    def __init__(self, item, event=None, performance=None, price_band_code=None,
                 price_band_description=None,
                 ticket_type_code=None, ticket_type_description=None,
                 ticket_orders=None, number_of_seats=None,
                 total_seatprice=None, total_surcharge=None,
                 seat_request_status=None, requested_seat_ids=None,
                 backend_purchase_reference=None, send_method=None,
                 gross_commission=None, user_commission=None):
        self.item = item
        self.event = event
        self.performance = performance
        self.price_band_code = price_band_code
        self.price_band_description = price_band_description
        self.ticket_type_code = ticket_type_code
        self.ticket_type_description = ticket_type_description
        self.ticket_orders = ticket_orders
        self.number_of_seats = number_of_seats
        self.total_seatprice = total_seatprice
        self.total_surcharge = total_surcharge
        self.seat_request_status = seat_request_status
        self.requested_seat_ids = requested_seat_ids
        self.backend_purchase_reference = backend_purchase_reference
        self.send_method = send_method
        self.gross_commission = gross_commission
        self.user_commission = user_commission

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **Order** object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a order.

        Returns:
            :class:`Order <pyticketswitch.order.Order>`: a new
            :class:`Order <pyticketswitch.order.Order>` object populated with
            the data from the api.

        """
        kwargs = {
            'item': data.get('item_number'),
            'number_of_seats': data.get('total_no_of_seats'),
            'price_band_code': data.get('price_band_code'),
            'price_band_description': data.get('price_band_desc'),
            'ticket_type_code': data.get('ticket_type_code'),
            'ticket_type_description': data.get('ticket_type_desc'),
            'seat_request_status': data.get('seat_request_status'),
            'requested_seat_ids': data.get('requested_seat_ids'),
            'backend_purchase_reference': data.get('backend_purchase_reference'),
        }

        raw_event = data.get('event')
        if raw_event:
            event = Event.from_api_data(raw_event)
            kwargs.update(event=event)

        raw_performance = data.get('performance')
        if raw_performance:
            performance = Performance.from_api_data(raw_performance)
            kwargs.update(performance=performance)

        raw_ticket_orders = data.get('ticket_orders', {}).get('ticket_order')
        if raw_ticket_orders:
            ticket_orders = [
                TicketOrder.from_api_data(ticket_order)
                for ticket_order in raw_ticket_orders
            ]
            kwargs.update(ticket_orders=ticket_orders)

        raw_total_seatprice = data.get('total_sale_seatprice')
        if raw_total_seatprice is not None:
            kwargs.update(total_seatprice=raw_total_seatprice)

        raw_total_surcharge = data.get('total_sale_surcharge')
        if raw_total_surcharge is not None:
            kwargs.update(total_surcharge=raw_total_surcharge)

        raw_send_method = data.get('send_method')
        if raw_send_method:
            send_method = SendMethod.from_api_data(raw_send_method)
            kwargs.update(send_method=send_method)

        raw_gross_commission = data.get('predicted_gross_commission')
        if raw_gross_commission is None:
            raw_gross_commission = data.get('gross_commission')
        if raw_gross_commission:
            gross_commission = Commission.from_api_data(raw_gross_commission)
            kwargs.update(gross_commission=gross_commission)

        raw_user_commission = data.get('predicted_user_commission')
        if raw_user_commission is None:
            raw_user_commission = data.get('user_commission')
        if raw_user_commission:
            user_commission = Commission.from_api_data(raw_user_commission)
            kwargs.update(user_commission=user_commission)

        return cls(**kwargs)

    def get_seats(self):
        """Get all seats from all :class:`TicketOrder <pyticketswitch.order.TicketOrder>` children.

        Returns:
            list: list of `Seats <pyticketswitch.seat.Seat>`.

        """
        if not self.ticket_orders:
            return []

        return [
            seat
            for ticket_order in self.ticket_orders
            if ticket_order.seats
            for seat in ticket_order.seats
        ]

    def get_seat_ids(self):
        """Get all seat ids from all :class:`TicketOrders <pyticketswitch.order.TickerOrder>`

        Returns:
            list: list of seat ids.

        """
        return [seat.id for seat in self.get_seats() if seat.id]

    def unique_seat_text(self):
        """Get the unique seat text for all seats in an order

        Returns:
            string: string combining all seat texts for the order

        """
        text_set = {
            seat.seat_text: seat.seat_text
            for seat in self.get_seats() if seat.seat_text
        }
        return ', '.join(seat_text for seat_text in text_set.values())

    def total_including_send_cost(self):
        return self.total_seatprice + self.total_surcharge + self.send_method.cost

    def __repr__(self):
        return u'<Order {}>'.format(self.item)



from pyticketswitch.mixins import JSONMixin
from pyticketswitch.utils import isostr_to_datetime


class Review(JSONMixin, object):
    """User or Critic evaluation of an event.

    Attributes:
        body (str): review test.
        date_time (datetime.datetime): date and time of the review.
        star_rating (int): rating on a scale of 1-5, with 1 being the lowest
            rating and 5 being the highest rating.
        language (str): the IETF language tag for the review.
        title (str): a review title if available.
        is_user (bool): the review was made by a user not a critic.
        author (str): the authors name.
        url (str): the original url.

    """

    def __init__(self, body=None, date_time=None, star_rating=None,
                 lang=None, title=None, is_user=False, author=None, url=None):

        self.body = body
        self.date_time = date_time
        self.star_rating = star_rating
        self.language = lang
        self.title = title
        self.is_user = is_user
        self.author = author
        self.url = url

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **Review** object from ticketswitch API data.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a review.

        Returns:
            :class:`Review <pyticketswitch.order.Review>`: a new
            :class:`Review <pyticketswitch.order.Review>` object
            populated with the data from the api.

        """

        review_date = isostr_to_datetime(
            data.get('review_iso8601_date_and_time')
        )

        kwargs = {
            'body': data.get('review_body', None),
            'date_time': review_date,
            'star_rating': data.get('star_rating', None),
            'lang': data.get('review_lang', None),
            'title': data.get('review_title', None),
            'is_user': data.get('is_user_review', False),
            'author': data.get('review_author', None),
            'url': data.get('review_original_url', None),
        }

        return cls(**kwargs)



from pyticketswitch.cost_range import CostRange
from pyticketswitch.discount import Discount
from pyticketswitch.seat import Seat, SeatBlock
from pyticketswitch.commission import Commission
from pyticketswitch.mixins import JSONMixin, SeatPricingMixin


class PriceBand(SeatPricingMixin, JSONMixin, object):
    """Describes a set of tickets with the same ticket type and price.

    Attributes:
        code (str): the price band identifier.
        default_discount (:class:`Discount <pyticketswitch.discount.Discount>`):
            this is the discount that will be assumed if no other discount
            is specified at reservation time. It holds the prices for the price
            band.
        allows_leaving_single_seats (str): indicates if this price band will
            allow customers to leave a single seat with no neighbours. Values
            are 'always', 'never', or 'if_necessary'.
        description (str): human readable description of the price band if
            available.
        cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            summary data for the price band including offers.
        no_singles_cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            summary data for the price band including offers.
        example_seats (list): list of :class:`Seats <pyticketswitch.seat.Seat>`
            that can be used as examples of what the user might get when they
            reserved tickets in this price band.
        example_seats_are_real (bool): when :obj:`True` this field indicates
            that the example seats are in fact real seats and will be the ones
            we attempt to reserve at the reservation stage. When :obj:`False`
            these seats merely examples retrieved from cached data, and have
            likely already been purchased.
        seat_blocks (list): list of
            :class:`SeatBlocks <pyticketswitch.seat.SeatBlock>`. When available
            this are the contiguous seats that are available for purchase.
            :class:`SeatBlocks <pyticketswitch.seat.SeatBlock>` contain
            :class:`Seats <pyticketswitch.seat.Seat>`.
        user_commission (:class:`Commission <pyticketswitch.commission.Commission>`):
            the commission payable to the user on the sale of tickets in this
            price band. Only available when requested.
        gross_commission (:class:`Commission <pyticketswitch.commission.Commission>`):
            the gross commission payable. This is not generally available.
        availability (int): the number of tickets/seats available in this price
            band.

    """

    def __init__(self, code, default_discount, description=None, cost_range=None,
                 no_singles_cost_range=None, example_seats=None,
                 example_seats_are_real=True, seat_blocks=None, user_commission=None,
                 discounts=None, allows_leaving_single_seats=None, availability=None,
                 seatprice=None, surcharge=None, non_offer_seatprice=None,
                 non_offer_surcharge=None, percentage_saving=0, absolute_saving=0):

        self.code = code
        self.description = description
        self.cost_range = cost_range
        self.allows_leaving_single_seats = allows_leaving_single_seats
        self.no_singles_cost_range = no_singles_cost_range
        self.default_discount = default_discount
        self.example_seats = example_seats
        self.example_seats_are_real = example_seats_are_real
        self.seat_blocks = seat_blocks
        self.user_commission = user_commission
        self.discounts = discounts
        self.availability = availability
        self.seatprice = seatprice
        self.surcharge = surcharge
        self.non_offer_seatprice = non_offer_seatprice
        self.non_offer_surcharge = non_offer_surcharge
        self.percentage_saving = percentage_saving
        self.absolute_saving = absolute_saving

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **PriceBand** object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a price band.

        Returns:
            :class:`PriceBand <pyticketswitch.order.PriceBand>`: a new
            :class:`PriceBand <pyticketswitch.order.PriceBand>` object
            populated with the data from the api.

        """
        api_cost_range = data.get('cost_range', {})
        api_no_singles_cost_range = api_cost_range.get('no_singles_cost_range', {})
        cost_range = None
        no_singles_cost_range = None

        if api_cost_range:
            api_cost_range['singles'] = True
            cost_range = CostRange.from_api_data(api_cost_range)

        if api_no_singles_cost_range:
            api_no_singles_cost_range['singles'] = False
            no_singles_cost_range = CostRange.from_api_data(
                api_no_singles_cost_range)

        discount = Discount.from_api_data(data)

        kwargs = {
            'code': data.get('price_band_code'),
            'description': data.get('price_band_desc'),
            'availability': data.get('number_available'),
            'cost_range': cost_range,
            'no_singles_cost_range': no_singles_cost_range,
            'default_discount': discount,
            'example_seats_are_real': data.get('example_seats_are_real', True),
            'allows_leaving_single_seats': data.get('allows_leaving_single_seats'),
        }

        example_seats_data = data.get('example_seats')
        if example_seats_data:
            example_seats = [
                Seat.from_api_data(seat)
                for seat in example_seats_data
            ]
            kwargs.update(example_seats=example_seats)

        seat_block_data = data.get('free_seat_blocks')

        if seat_block_data:
            separators_by_row = seat_block_data.get('separators_by_row')
            restricted_view_seats = seat_block_data.get('restricted_view_seats')
            seats_by_text_message = seat_block_data.get('seats_by_text_message')
            blocks_by_row = seat_block_data.get('blocks_by_row')

            seat_blocks = []
            if blocks_by_row:
                for row_id, row in blocks_by_row.items():
                    for block in row:
                        separator = separators_by_row.get(row_id)
                        seat_block = SeatBlock.from_api_data(
                            block=block,
                            row_id=row_id,
                            separator=separator,
                            restricted_view_seats=restricted_view_seats,
                            seats_by_text_message=seats_by_text_message,
                        )
                        seat_blocks.append(seat_block)

            kwargs.update(seat_blocks=seat_blocks)

        user_commission_data = data.get('predicted_user_commission')
        if user_commission_data:
            user_commission = Commission.from_api_data(user_commission_data)
            kwargs.update(user_commission=user_commission)

        discounts_data = data.get('possible_discounts', {}).get('discount')
        if discounts_data:
            discounts = [
                Discount.from_api_data(discount_data)
                for discount_data in discounts_data
            ]
            kwargs.update(discounts=discounts)

        kwargs.update(SeatPricingMixin.kwargs_from_api_data(data))

        return cls(**kwargs)

    def get_seats(self):
        """Get all seats in child seat blocks

        Returns:
            list: list of :class:`Seats <pyticketswitch.seat.Seat>`.

        """
        if not self.seat_blocks:
            return []

        return [
            seat
            for seat_block in self.seat_blocks
            for seat in seat_block.seats or []
        ]

    def __repr__(self):
        if self.description:
            return u'<PriceBand {}:{}>'.format(
                self.code, self.description.encode('ascii', 'ignore'))
        return u'<PriceBand {}>'.format(self.code)



from pyticketswitch.mixins import JSONMixin


class Content(JSONMixin, object):
    """Plain text and HTML content

    Attributes:
        name (str): the identifier of the content
        value (str): plain text content.
        value_html (str): html content.

    """

    def __init__(self, name=None, value=None, value_html=None):

        self.name = name
        self.value = value
        self.value_html = value_html

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Content object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns some content..

        Returns:
            :class:`Content <pyticketswitch.content.Content>`: a new
            :class:`Content <pyticketswitch.content.Content>` object
            populated with the data from the api.

        """

        kwargs = {
            'name': data.get('name'),
            'value': data.get('value'),
            'value_html': data.get('value_html'),
        }

        return cls(**kwargs)



from pyticketswitch.mixins import JSONMixin, SeatPricingMixin
from pyticketswitch.commission import Commission


class Discount(SeatPricingMixin, JSONMixin, object):
    """Represents a set of prices for a price band.

    Attributes:
        code (str): the identifier for the discount.
        description (str): a human readable description for the discount.
        price_band_code (str): identifier for the related price band.
        is_offer (bool): indicates that the discount is an offer.
        availability (int): the number tickets available with this discount.
        percentage_saving (float): the amount saved compared to when this
            discount is not on offer, as a percentage or the non offer price.
        absolute_saving (float): the amount saved compared to when this
            discount is not on offer.
        gross_commission (:class: `Commission <pyticketswitch.commission.Commission>`):
            predicted commission to be shared between ingresso and the partner.
        user_commission (:class:`Commission <pyticketswitch.commission.Commission>`):
            predicted commission for the partner.
        disallowed_seat_nos (list): a list of seat numbers, that this discount
            code cannot be specified for.
    """

    def __init__(self, code, description=None, price_band_code=None,
                 availability=None, is_offer=False, percentage_saving=0,
                 absolute_saving=0, gross_commission=None, user_commission=None,
                 disallowed_seat_nos=None, *args, **kwargs):
        super(Discount, self).__init__(*args, **kwargs)
        self.code = code
        self.description = description
        self.price_band_code = price_band_code
        self.is_offer = is_offer
        self.availability = availability
        self.percentage_saving = percentage_saving
        self.absolute_saving = absolute_saving
        self.gross_commission = gross_commission
        self.user_commission = user_commission
        self.disallowed_seat_nos = disallowed_seat_nos

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Discount object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a discount.

        Returns:
            :class:`Discount <pyticketswitch.discount.Discount>`: a new
            :class:`Discount <pyticketswitch.discount.Discount>` object
            populated with the data from the api.

        """
        gross_commission = data.get('predicted_gross_commission')
        if gross_commission:
            gross_commission = Commission.from_api_data(gross_commission)

        user_commission = data.get('predicted_user_commission')
        if user_commission:
            user_commission = Commission.from_api_data(user_commission)

        kwargs = {
            'code': data.get('discount_code'),
            'description': data.get('discount_desc'),
            'disallowed_seat_nos': data.get('discount_disallowed_seat_nos'),
            'price_band_code': data.get('price_band_code'),
            'is_offer': data.get('is_offer', False),
            'seatprice': data.get('sale_seatprice'),
            'surcharge': data.get('sale_surcharge'),
            'non_offer_seatprice': data.get('non_offer_sale_seatprice'),
            'non_offer_surcharge': data.get('non_offer_sale_surcharge'),
            'availability': data.get('number_available'),
            'percentage_saving': data.get('percentage_saving'),
            'absolute_saving': data.get('absolute_saving'),
            'gross_commission': gross_commission,
            'user_commission': user_commission,
        }
        kwargs.update(SeatPricingMixin.kwargs_from_api_data(data))

        return cls(**kwargs)

    def __repr__(self):
        return u'<Discount {}:{}>'.format(
            self.code, self.description.encode('ascii', 'ignore'))



from pyticketswitch.mixins import JSONMixin


class SeatBlock(JSONMixin, object):
    """Describes a set of contiguous seats.

    Attributes:
        length (int): the number of seats in the block.
        seats (list): list of :class:`Seats <pyticketswitch.seat.Seat>` in the
            block.

    """

    def __init__(self, length, seats=None):
        self.length = length
        self.seats = seats

    @classmethod
    def from_api_data(cls, block, row_id=None, separator='',
                      restricted_view_seats=None, seats_by_text_message=None):
        """Creates a new SeatBlock object from API data from ticketswitch.

        Args:
            block (list): the part of the response from a ticketswitch API call
                that concerns a seat block.
            row_id (str): the component of the seat ID corresponding to the row
            separator (str): the string separating row and column in the id
            restricted_view_seats (list): seat IDs that have restricted view
            seats_by_text_message (dict): a mapping of seat text messages to
                seat IDs

        Returns:
            :class:`SeatBlock <pyticketswitch.seat.SeatBlock>`: a new
            :class:`SeatBlock <pyticketswitch.seat.SeatBlock>` object
            populated with the data from the api.

        """

        seats = []
        for seat_id in block:

            column = None
            delimiter = separator

            if not delimiter:
                delimiter = row_id

            split_id = seat_id.split(delimiter)
            column = split_id[1]

            restricted = False
            restricted_text = ''

            if seat_id in restricted_view_seats:
                restricted = True

            for seat_text, list_of_seats in seats_by_text_message.items():
                if seat_id in list_of_seats:
                    restricted_text = seat_text
            seat = Seat(
                id_=seat_id, row=row_id, column=column,
                separator=separator, is_restricted=restricted,
                seat_text=restricted_text
            )
            seats.append(seat)

        kwargs = {'seats': seats, 'length': len(seats)}
        return cls(**kwargs)


class Seat(JSONMixin, object):
    """Describes a seat in a venue.

    Attributes:
        id (str): the identifier for the seat.
        column (str): the column of the seat.
        row (str): the row of the seat.
        separator (str): characters that should be used to seperate the column
            and row when presenting seat information.
        is_restricted (bool): indicates that the seat has a restricted view.
        seat_text_code (str): code indicating text that should be displayed
            with the seat when preseting seat information.
        seat_text (str): readable explanation of the seats description
        barcode (str): barcode specific to this seat. Only available when
            supported by the backend system.

    """

    def __init__(self, id_=None, column=None, row=None, is_restricted=False,
                 seat_text_code=None, seat_text=None, separator=None,
                 barcode=None):
        self.id = id_
        self.column = column
        self.row = row
        self.separator = separator
        self.is_restricted = is_restricted
        self.seat_text = seat_text
        self.seat_text_code = seat_text_code
        self.barcode = barcode

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Seat object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a seat.

        Returns:
            :class:`Seat <pyticketswitch.seat.Seat>`: a new
            :class:`Seat <pyticketswitch.seat.Seat>` object
            populated with the data from the api.

        """
        kwargs = {
            'id_': data.get('full_id'),
            'column': data.get('col_id'),
            'row': data.get('row_id'),
            'is_restricted': data.get('is_restricted_view', False),
            'seat_text_code': data.get('seat_text_code'),
            'seat_text': data.get('seat_text'),
            'separator': data.get('separator', ''),
            'barcode': data.get('barcode'),
        }

        return cls(**kwargs)

    def __repr__(self):
        return u'<Seat {}>'.format(self.id)



from pyticketswitch.mixins import JSONMixin


class Debitor(JSONMixin, object):
    """Information about a 3rd party that will take payment from your customer.

    This information is primarily used for bypassing callouts an integrating
    directly with payment providers on the front end.

    When your account is set up to sell on credit (i.e. you are always taking
    payment from the customer in your application directly), this information
    will not be present, and it should not be relevant.

    When the source system is taking payment this information will not be
    present.

    When debitor information is not present or you are not front end
    integrating you should refer to the
    :attr:`Reservation.needs_payment_card
    <pyticketswitch.reservation.Reservation.needs_payment_card>`,
    :attr:`Reservation.needs_email_address
    <pyticketswitch.reservation.Reservation.needs_email_address>`, and
    :attr:`Reservation.needs_agent_reference
    <pyticketswitch.reservation.Reservation.needs_agent_reference>` attributes
    as to what information you need to pass back to the API for purchasing
    tickets.

    Regardless of the debitor it's advisable to implement the full
    purchase/callout/callback process in the event that your front end
    integration goes awry.

    :ref:`See front end integrations for more information
    <frontend_integrations>`

    Attributes:
        type (str): all debitors with the same type can be assumed to integrate
            in the same manor, however their parameters and integration data
            might be different.
        name (str): name of the specific implementation of the debitor.
        description (str): human readable description of the debitor.
        integration_data (dict): data used to do front end integrations. For
            the format of this data please consult :ref:`the documentation for
            the relevant debitor type <frontend_integrations>`.
        aggregation_key (str): a key used to identify if debitors are the same
            for purposes of payment aggregation between bundles
    """

    def __init__(self, typ=None, name=None, description=None,
                 integration_data=None, aggregation_key=None):
        self.type = typ
        self.name = name
        self.description = description
        self.integration_data = integration_data
        self.aggregation_key = aggregation_key

    @classmethod
    def from_api_data(cls, data):
        kwargs = {
            'typ': data.get('debitor_type'),
            'name': data.get('debitor_name'),
            'description': data.get('debitor_desc'),
            'integration_data': data.get('debitor_integration_data', {}),
            'aggregation_key': data.get('debitor_aggregation_key'),
        }

        return cls(**kwargs)

    def __repr__(self):
        return u'<Debitor {}:{}>'.format(self.type, self.name)



from pyticketswitch.mixins import JSONMixin


class Address(JSONMixin, object):
    """A postal address.

    TODO: this is very similar the Customer object. consider combining them,
    or subclassing one from the other, or a base class of some sort.

    Attributes:
        lines (list): list of address lines.
        town (str): town for the address.
        county (str): the county or region of the address.
        country_code (str): ISO 3166-1 country code.
        post_code (str): post or ZIP code for the address.
        email (str): a contact email address.
        home_phone (str): a contact home phone number.
        work_phone (str): a contact work phone number.

    """

    def __init__(self, lines=None, country_code=None, county=None,
                 email=None, home_phone=None, post_code=None, town=None,
                 work_phone=None):

        self.lines = lines
        self.town = town
        self.county = county
        self.country_code = country_code
        self.post_code = post_code
        self.email = email
        self.home_phone = home_phone
        self.work_phone = work_phone

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Address object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns an address.

        Returns:
            :class:`Address <pyticketswitch.address.Address>`: a new
            :class:`Address <pyticketswitch.address.Address>` object
            populated with the data from the api.

        """
        kwargs = {
            'country_code': data.get('country_code'),
            'county': data.get('county'),
            'email': data.get('email_address'),
            'home_phone': data.get('home_phone'),
            'post_code': data.get('postcode'),
            'town': data.get('town'),
            'work_phone': data.get('work_phone'),
            'lines': [
                data.get('address_line_one'),
                data.get('address_line_two'),
            ]
        }

        return cls(**kwargs)

    def as_api_billing_address_parameters(self):

        params = {
            'billing_country_code': self.country_code,
            'billing_postcode': self.post_code,
            'billing_town': self.town,
            'billing_county': self.county,
        }

        if self.lines and len(self.lines) == 1:
            params.update(billing_address_line_one=self.lines[0])

        if self.lines and len(self.lines) > 1:
            params.update(
                billing_address_line_one=self.lines[0],
                billing_address_line_two=self.lines[1],
            )
        return params



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='pyticketswitch',
    version='2.7.4',
    author='Ingresso',
    author_email='systems@ingresso.co.uk',
    url='https://github.com/ingresso-group/pyticketswitch/',
    packages=[
        'pyticketswitch',
    ],
    license='MIT',
    description='A Python interface for the Ingresso F13 API',
    install_requires=[
        'requests>=2.0.0',
        'python-dateutil>2.5.3',
        'six>=1.11.0',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: Developers',
        'Natural Language :: English',
    ],
)



from pyticketswitch.bundle import Bundle
from pyticketswitch.order import Order
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.purchase_result import PurchaseResult


class Trolley(JSONMixin, object):
    """Describes a collection of products that are being purchased.

    Attributes:
        token (str): a token that represents the current state of the trolley.
        transaction_uuid (str): unique identifier for this trolley. Only
            present after reservation.
        transaction_id (str): unique identifier for this trolley. Only present
            after purchase. This supposed to be an identifer that is easily
            readable/recognisable by a human being.
        bundles (list): list of :class:`Bundles <pyticketswitch.bundle.Bundle>`
            objects that break down the items to be purchased by their source
            systems.
        discarded_orders (list): list of
            :class:`Orders <pyticketswitch.order.Order>` objects that were in
            the trolley in the past but have been removed.
        minutes_left (float): the number of minutes left before a reservation
            expires.
        order_count (int): the number of orders in the trolley.
        purchase_result (:class:`PurchaseResult <pyticketswitch.callout.Callout>`):
            the result of the purchase attempt when available.
        input_contained_unavailable_order (bool): indicates that the call used
            to create or modify this trolley object included at least one order
            that was not available.
    """
    def __init__(self, token=None, transaction_uuid=None, transaction_id=None,
                 bundles=None, discarded_orders=None, minutes_left=None,
                 order_count=None, purchase_result=None,
                 input_contained_unavailable_order=False):
        self.token = token
        self.transaction_uuid = transaction_uuid
        self.transaction_id = transaction_id
        self.bundles = bundles
        self.discarded_orders = discarded_orders
        self.minutes_left = minutes_left
        self.order_count = order_count
        self.purchase_result = purchase_result
        self.input_contained_unavailable_order = input_contained_unavailable_order

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Trolley object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a trolley.

        Returns:
            :class:`Trolley <pyticketswitch.trolley.Trolley>`: a new
            :class:`Trolley <pyticketswitch.trolley.Trolley>` object
            populated with the data from the api.

        """
        raw_contents = data.get('trolley_contents', {})

        if not raw_contents:
            raw_contents = data.get('reserved_trolley', {})

        if not raw_contents:
            raw_contents = data.get('trolley_token_contents', {})

        raw_bundles = raw_contents.get('bundle', [])

        bundles = [
            Bundle.from_api_data(bundle)
            for bundle in raw_bundles
        ]

        raw_discarded_orders = data.get('discarded_orders', [])

        discarded_orders = [
            Order.from_api_data(order)
            for order in raw_discarded_orders
        ]

        kwargs = {
            'token': data.get('trolley_token'),
            'bundles': bundles,
            'discarded_orders': discarded_orders,
            'transaction_uuid': raw_contents.get('transaction_uuid'),
            'transaction_id': raw_contents.get('transaction_id'),
            'order_count': data.get('trolley_order_count'),
            'input_contained_unavailable_order': data.get(
                'input_contained_unavailable_order', False),
        }

        minutes = data.get('minutes_left_on_reserve')
        if minutes is not None:
            kwargs.update(minutes_left=float(minutes))

        purchase_result = raw_contents.get('purchase_result')
        if purchase_result:
            kwargs.update(
                purchase_result=PurchaseResult.from_api_data(purchase_result))

        return cls(**kwargs)

    def get_events(self):
        """Get the events in the trolley.

        Returns:
            list: list of :class:`Event <pyticketswitch.event.Event>` objects.
        """
        if not self.bundles:
            return []
        return [
            event
            for bundle in self.bundles
            for event in bundle.get_events()
            if event and event.id
        ]

    def get_event_ids(self):
        """Get the event ids of the events in the trolley.

        Returns:
            set: set of events IDs (str).

        """
        return {event.id for event in self.get_events()}

    def get_bundle(self, source_code):
        """Find a bundle with a specific source code

        Args:
            source_code (str): the source code for the target bundle.

        Returns:
            :class:`Bundle <pyticketswitch.bundle.Bundle>`: the bundle with the
            requested source code.

            When no bundle with that source code is present returns :obj:`None`.

        """

        if not self.bundles:
            return None

        for bundle in self.bundles:
            if bundle.source_code == source_code:
                return bundle

        return None

    def get_item(self, item_number):
        """Find an order with the given item number

        Args:
            item_number (int): the item number of the target order

        Returns:
            :class:`Order <pyticketswitch.order.order>`: the order with the
            specified

        """
        if not self.bundles:
            return None

        for order in self.get_orders():
            if order.item == item_number:
                return order

        return None

    def get_orders(self):
        """Get all orders in all bundles in a trolley

        Returns:
            list: orders from all bundles in the the trolley

        """

        if not self.bundles:
            return []

        return [
            order
            for bundle in self.bundles
            if bundle.orders
            for order in bundle.orders
        ]

    def __repr__(self):
        if self.transaction_id:
            return u'<Trolley id:{}>'.format(self.transaction_id)
        if self.transaction_uuid:
            return u'<Trolley uuid:{}>'.format(self.transaction_uuid)
        if self.token:
            return u'<Trolley token:{}>'.format(self.token)

        return super(Trolley, self).__repr__()



from pyticketswitch.price_band import PriceBand
from pyticketswitch.mixins import JSONMixin


class TicketType(JSONMixin, object):
    """Describes a collection of tickets.

    Generally this represents a part of house in a venue, but may have other
    meanings in contexts outside theater and music.

    Attributes:
        code (str): identifier for the ticket type.
        description (str): human readable description of the ticket type.
        price_bands (list): list of
            :class:`PriceBands <pyticketswitch.price_band.PriceBand>` objects
            wich further subdivided available tickets/seats by price.

    """
    def __init__(self, code=None, description=None, price_bands=None):

        self.code = code
        self.description = description
        self.price_bands = price_bands

    @classmethod
    def from_api_data(cls, data):
        """Creates a new PriceBand object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a price band.

        Returns:
            :class:`PriceBand <pyticketswitch.price_band.PriceBand>`: a new
            :class:`PriceBand <pyticketswitch.price_band.PriceBand>` object
            populated with the data from the api.

        """
        price_bands = []
        api_price_bands = data.get('price_band', [])
        for single_band in api_price_bands:
            price_bands.append(PriceBand.from_api_data(single_band))

        kwargs = {
            'code': data.get('ticket_type_code', None),
            'description': data.get('ticket_type_desc', None),
            'price_bands': price_bands,
        }

        return cls(**kwargs)

    def get_seats(self):
        """Get seats in the ticket type.

        Returns:
            list: list of :class:`Seats <pyticketswitch.seat.Seat>` objects.

        """
        if not self.price_bands:
            return []

        return [
            seat
            for price_band in self.price_bands
            for seat in price_band.get_seats()
        ]

    def __repr__(self):
        return u'<TicketType {}: {}>'.format(
            self.code, self.description.encode('ascii', 'ignore'))



import six
from abc import ABCMeta
from pyticketswitch.exceptions import InvalidParametersError


@six.add_metaclass(ABCMeta)
class PaymentMethod(object):
    """Abstract base class for payment methods"""

    def as_api_parameters(self):
        """Generate keyword arguments suitable for consumption by the
        ticketswitch API

        Returns:
            dict: dictionary of keyword parameters to pass to the API call.
        """
        raise NotImplementedError(
            'as_api_parameters not implemented on ' + self.__class__)


class CardDetails(object):
    """Credit card details

    This should never be returned by the API and is only used for supplying
    card details to the purchase call.

    Implements
    :class:`PaymentMethod <pyticketswitch.payment_methods.PaymentMethod>`.

    Attributes:
        card_number (str): the long credit card number.
        expiry_month (int): the month the card expires in. Defaults to
            :obj:`None`.
        expiry_year (int): the year the card expires in. defaults to
            :obj:`None`.
        start_month (int): the month the card expires in. Defaults to
            :obj:`None`.
        start_year (int): the year the card expires in. :obj:`None`.
        ccv2 (str): credit card security code. Defaults to :obj:`None`.
        issue_number (str): issue number of the card. Defaults to :obj:`None`.
        billing_address (:class:`Address <pyticketswitch.address.Address>`):
            used when the customer wishes to use an alternate billing address.
            when not specified the customer address will be used.
        return_url (str): some card debitors may decide that they need to
            redirect the user to a third party for verification (for example
            3d secure). These parties need a location to return a customer to.
            When available, it's recomended to provide it. In the situation
            where a return url is required but not provided, then the payment
            will fail.
        return_token (str): a unique token that can be used by you to identify
            when a card debitor returns to you.
        user_agent (str): the customer's browser's User-Agent header.
            only nessicary when providing a return url.
        accept (str): the customer's browser's Accept header.
        remote_site (str): the remote site's domain. must match the domain of
            the return_url.
    """

    def __init__(self, card_number, expiry_month=None,
                 expiry_year=None, start_month=None, start_year=None,
                 ccv2=None, issue_number=None, billing_address=None,
                 return_url=None, return_token=None, user_agent=None,
                 accept=None, remote_site=None):

        self.card_number = card_number
        self.expiry_month = expiry_month
        self.expiry_year = expiry_year
        self.start_month = start_month
        self.start_year = start_year
        self.ccv2 = ccv2
        self.issue_number = issue_number
        self.billing_address = billing_address
        self.return_url = return_url
        self.return_token = return_token
        self.user_agent = user_agent
        self.accept = accept
        self.remote_site = remote_site

    def as_api_parameters(self):
        """Generates a dictionary of parameters to be passed back to the API.

        Returns:
            dict: a set of parameters describing the card details to the API.

        """
        params = {
            'card_number': self.card_number,
        }

        missing_expiry_year = not self.expiry_year
        missing_expiry_month = not self.expiry_month

        if missing_expiry_year or missing_expiry_month:
            raise InvalidParametersError(
                'both expiry_year and expiry_month must be specified')

        params.update(
            expiry_date='{:0>2}{:0>2}'.format(
                self.expiry_month,
                # handle 4 digit years
                str(self.expiry_year)[-2:]
            )
        )

        missing_start_year = not self.start_year
        missing_start_month = not self.start_month

        specifying_start_date = self.start_year or self.start_month

        if specifying_start_date and (missing_start_year or missing_start_month):
            raise InvalidParametersError(
                'both start_year and start_month must be specified or neither specified')

        if specifying_start_date:
            params.update(
                start_date='{:0>2}{:0>2}'.format(
                    self.start_month,
                    str(self.start_year)[-2:]
                )
            )

        if self.ccv2:
            params.update(cv_two=self.ccv2)

        if self.issue_number:
            params.update(issue_number=self.issue_number)

        if self.billing_address:
            params.update(
                **self.billing_address.as_api_billing_address_parameters()
            )

        if self.return_url:
            params.update(return_url=self.return_url)

        if self.return_token:
            params.update(return_token=self.return_token)

        if self.user_agent:
            params.update(client_http_user_agent=self.user_agent)

        if self.accept:
            params.update(client_http_accept=self.accept)

        if self.remote_site:
            params.update(remote_site=self.remote_site)

        return params


class RedirectionDetails(object):
    """Information that specifies where a customer will be returned to after
    being redirected to an external payment provider.

    Implements
    :class:`PaymentMethod <pyticketswitch.payment_methods.PaymentMethod>`.

    Attributes:
        token (str): a unique token that can be used by you to identify when
            this callout returns to your website.
        url (str): the URL that the payment provider should redirect back to on
            success/failure of the customers payment.
        user_agent (str): the customer's browser's User-Agent header.
        accept (str): the customer's browser's Accept header.
        remote_site (str): the remote site's domain must match the domain of the
            return_url.

    """

    def __init__(self, token, url, user_agent, accept, remote_site):
        self.token = token
        self.url = url
        self.user_agent = user_agent
        self.accept = accept
        self.remote_site = remote_site

    def as_api_parameters(self):
        """Generate API keyword args for these details.

        Returns:
            dict: the redirection details in a format the API will understand.

        """
        return {
            'return_token': self.token,
            'return_url': self.url,
            'client_http_user_agent': self.user_agent,
            'client_http_accept': self.accept,
            'remote_site': self.remote_site,
        }


class StripeDetails(object):
    """For use with self generated stripe tokens

    Can be used to provide stripe tokens directly to the API at purchase time
    avoiding a callout/callback cycle.

    Implements
    :class:`PaymentMethod <pyticketswitch.payment_methods.PaymentMethod>`.

    Attributes:
        tokens (dict): dictionary of stripe card tokens indexed on bundle source
            code. If there are multiple bundles in the trolley, then a unique
            stripe token must be provided for each of the bundles you wish to
            purchase with stripe.
    """

    def __init__(self, tokens):
        self.tokens = tokens

    def as_api_parameters(self):
        """Generate API keyword args for these details.

        Returns:
            dict: the stripe details in a format the API will understand.

        """

        return {
            '{}_callback/stripeToken'.format(source): token
            for source, token in self.tokens.items()
        }


class CiderDetails(object):
    """For use with multiple payment tokens and details

    Can be used to provide payment tokens and details directly to the API at
    purchase time, avoiding a callout/callback cycle.

    Implements
    :class:`PaymentMethod <pyticketswitch.payment_methods.PaymentMethod>`.

    Attributes:
        data (dict): dictionary of payment tokens and details
        system_codes (list): a list of systems for which payment is being made
    """

    def __init__(self, data, system_codes):
        self.data = data
        self.system_codes = system_codes

    def as_api_parameters(self):
        """Generate API keyword args for these details.

        Returns:
            dict: the cider debitor details in the format the API can use.
        """

        data = {}
        for system in self.system_codes:
            data.update({
                "{0}_callback/{1}".format(system, variable): self.data[variable]
                for variable in self.data.keys()
            })
        return data


PaymentMethod.register(CardDetails)
PaymentMethod.register(RedirectionDetails)
PaymentMethod.register(StripeDetails)
PaymentMethod.register(CiderDetails)



from pyticketswitch.mixins import JSONMixin


class User(JSONMixin, object):
    """Describes a user of the API

    Attributes:
        id (str): the user identifier.
        name (str): human readable name.
        country (str): ISO 3166-1 country code.
        sub_user (str): the identifier of the sub user.
        is_b2b (bool): indicates that the account is a b2b account.
        statement_descriptor (str): what will appear on a customers bank
            statement when ingresso takes the payment.
        backend_group (str): what product group a user belongs to. Users in
            the same backend_group will see the same products.
        content_group (str): what content group a user belongs to. Users in
            the same content_group will see the same textual and graphical
            content.

    """
    def __init__(self, id_, name=None, country=None, sub_user=None,
                 is_b2b=False, statement_descriptor=None,
                 backend_group=None, content_group=None):

        self.id = id_
        self.name = name
        self.country = country
        self.sub_user = sub_user
        self.is_b2b = is_b2b
        self.statement_descriptor = statement_descriptor
        self.backend_group = backend_group
        self.content_group = content_group

    @classmethod
    def from_api_data(cls, data):
        """Creates a new User object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a user.

        Returns:
            :class:`User <pyticketswitch.user.User>`: a new
            :class:`User <pyticketswitch.user.User>` object
            populated with the data from the api.

        """
        return cls(
            data.get('user_id'),
            name=data.get('real_name'),
            country=data.get('default_country_code'),
            sub_user=data.get('sub_user'),
            is_b2b=data.get('is_b2b'),
            statement_descriptor=data.get('statement_descriptor'),
            backend_group=data.get('backend_group'),
            content_group=data.get('content_group'),
        )



import datetime

from pyticketswitch.misc import MONTH_NUMBERS
from pyticketswitch.cost_range import CostRange
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.utils import bitmask_to_numbered_list


class Month(JSONMixin, object):
    """Availability summary data for a given month

    Attributes:
        month (int): number of the month, jan == 1, feb == 2, dec == 12, etc.
        year (int): year of the month. eg. 2014, 2015, 2016.
        description (str): the human readable description of the month.
        cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricing summary, may also include offers.
        no_singles_cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricing summary when no leaving single seats, may also include
            offers.

    """
    def __init__(self, month, year, description=None, dates_bitmask=None,
                 weekday_bitmask=None, cost_range=None,
                 no_singles_cost_range=None):

        self.month = month
        self.year = year
        self.description = description
        self._dates_bitmask = dates_bitmask
        self._weekday_bitmask = weekday_bitmask
        self.cost_range = cost_range
        self.no_singles_cost_range = no_singles_cost_range

    @classmethod
    def from_api_data(cls, data):

        api_cost_range = data.get('cost_range', {})
        api_no_singles_cost_range = api_cost_range.get(
            'no_singles_cost_range', {})
        cost_range = None
        no_singles_cost_range = None

        if api_cost_range:
            cost_range = CostRange.from_api_data(api_cost_range)

        if api_no_singles_cost_range:
            no_singles_cost_range = CostRange.from_api_data(
                api_no_singles_cost_range)

        kwargs = {
            'cost_range': cost_range,
            'no_singles_cost_range': no_singles_cost_range,
            'month': MONTH_NUMBERS.get(data.get('month')),
            'year': data.get('year'),
            'description': data.get('month_desc'),
            'dates_bitmask': data.get('month_dates_bitmask'),
            'weekday_bitmask': data.get('month_weekdays_bitmask'),
        }

        return cls(**kwargs)

    def is_available(self, day):
        """Indicates if the event has availability on a given day.

        the dates bitmask is a 32 bit int where the right most bit is the first
        day of the month. To find if we have availability on a specfic day,
        we bit shift the mask by the zero indexed day and AND it by 1. If the
        result is a 1 then this combo of ticket type and price band is
        available on that day, otherwise it is not.

        Args:
            day (int): the day of the month.

        Returns:
            bool: :obj:`True` there is availability on the given day. Otherwise
            :obj:`False`.
        """

        return bool(self._dates_bitmask >> (day - 1) & 1)

    def on_weekday(self, day):
        """Indicates if the event has availability on a given day of the week.

        Check if the month has shows available on a given day of the week,
        where 0 is monday and sunday is 6.

        .. note:: the api uses sunday as day 0 where as python uses monday.
                  I've made a concious decision here to use the python numbers
                  to keep it consistent with anything written against this.
                  (also a sunday day 0 is silly).

        FIXME: this is copied directly from the AvailabilityDetails class.
        Perhaps worth breaking this out to common base or mixin class?

        Args:
            day (int): the day of the week. monday = 0, tuesday = 1, sunday = 6.

        Returns:
            bool: :obj:`True` there is availability on the given day. Otherwise
            :obj:`False`.

        """
        adjusted_day = day + 1 if day < 6 else 0
        return bool(self._weekday_bitmask >> adjusted_day & 1)

    def start_date(self):
        num_list = bitmask_to_numbered_list(self._dates_bitmask)
        if not num_list:
            return
        first_day = num_list[0]
        howdy = datetime.date(year=self.year, month=self.month, day=first_day)
        return howdy

    def end_date(self):
        num_list = bitmask_to_numbered_list(self._dates_bitmask)
        if not num_list:
            return
        last_day = num_list[-1]
        howdy = datetime.date(year=self.year, month=self.month, day=last_day)
        return howdy
